(***********************************************************************
 * File: State_valid.v
 * $Revision: 455 $
 * $Date: 2011-01-25 20:29:37 -0200 (Tue, 25 Jan 2011) $
 * $LastChangedBy: jdcampo $
 ***********************************************************************
 * This file provides mild conditions under which hypervisor calls terminate
 **********************************************************************)

Require Export State_proof.
Require Export State_effects.
Require Export Exec.

Section Prop_hcalls.

Variable ctxt : context.

(* no hypercall is pending in a state *)
Definition noHcall (s:state) : Prop :=
  forall osi:os_ident,
    exists2 tos:os, 
      Value os_ident tos = map_apply os_ident_eq (st_oss s) osi &
      os_hcall tos = None.


(* if the hypervisor is not running, then no hypercall is pending *)
Definition osRunningNoHcall (s:state) : Prop :=
  ~(hyper_running s) -> noHcall s.


(* Auxiliar lemmas *)

Lemma osRunningNoHcall_Chmod : 
  forall (s s':state),
  noHcall s ->
  valid_exec ctxt s Chmod s' ->
  osRunningNoHcall s'.
Proof.
  intros s s'; unfold osRunningNoHcall, noHcall; intros not_hcall_s valid_exec_chmod. 
  inversion_clear valid_exec_chmod as [s_vexec s'_vexec act_vexec H_pre H_post].
  unfold post_action, post_chmod in H_post.
  destruct H_post as [(Htrusted, Hassign) | (Huntrusted, Hassign)].
  unfold state_components_assign, state_components_update in Hassign.
  destruct Hassign with Guest_OSs as [(Heqcomp1 , Hinterp) | (Heqcomp2, Hinterp)].
  unfold state_component_interp in Hinterp; rewrite <- Hinterp; auto.
  elim Heqcomp2; split; discriminate.
  unfold state_components_assign, state_components_update in Hassign.
  destruct Hassign with Guest_OSs as [(Heqcomp1 , Hinterp) | (Heqcomp2, Hinterp)].
  unfold state_component_interp in Hinterp; rewrite <- Hinterp; auto.
  elim Heqcomp2; split; discriminate.
Qed.


Lemma post_Chmod_os_running : 
             forall(s s' : state), 
             valid_exec ctxt s Chmod s' ->                       
             ~ hyper_running s'.
Proof.
  unfold not, hyper_running; intros s s' valide hyperr.
  inversion_clear valide as [s_vexec s'_vexec act_vexec H_pre H_post].
  unfold post_action, post_chmod in H_post.
  destruct H_post as [(Htrusted, Hassign) | (Huntrusted, Hassign)].
  unfold state_components_assign, state_components_update in Hassign.
  destruct Hassign with OS_Activity as [((eq1, eq2), Hinterp) | (noteq, Hinterp)]; try discriminate.
  elim eq2; auto.
  unfold state_component_interp in Hinterp; destruct Hinterp as (Hexecm, Hosact).
  rewrite Hosact in hyperr; discriminate.
  unfold state_components_assign, state_components_update in Hassign.
  destruct Hassign with OS_Activity as [((eq1, eq2), Hinterp) | (noteq, Hinterp)]; try discriminate.
  elim eq2; auto.
  unfold state_component_interp in Hinterp; destruct Hinterp as (Hexecm, Hosact).
  rewrite Hosact in hyperr; discriminate.
Qed.


(* The only action that produces a hypercall is Hcall *)
Lemma noHcall_inv : 
  forall (s s':state) (a: action), 
  (forall (hc: Hyperv_call), a <> Hcall hc) ->
  noHcall s ->
  valid_exec ctxt s a s'->
  noHcall s'.
Proof.
  unfold noHcall; intros s s' a a_not_Hcall no_hcall_s valide.
  induction a; inversion_clear valide as [s_vexec s'_vexec act_vexec H_pre H_post]; unfold pre_action in H_pre; unfold post_action in H_post.

  (* Silent *)
  unfold post_silent, same_state in H_post.
  rewrite <- H_post; auto.

  (* Read *)
  unfold post_read, post_all_read, same_state in H_post.
  rewrite <- H_post; auto.

  (* Read_hyper *)
  unfold post_read_hyper, post_all_read, same_state in H_post.
  rewrite <- H_post; auto.
 
 (* Write *)
  unfold post_write, post_all_write in H_post.
  destruct H_post as (ma, va_ma, (newm, Hupd, Hassign)).
  unfold assign_system_memory, state_component_assign, state_component_update in Hassign.
  destruct Hassign with Guest_OSs as [(Heq, Hinterp) | (Hnoteq, Hinterp)]; try discriminate.
  unfold state_component_interp in Hinterp; rewrite <- Hinterp; auto.

  (* Write_hyper *)
  unfold post_write_hyper, post_all_write in H_post.
  destruct H_post as (ma, va_ma, (newm, Hupd, Hassign)).
  unfold assign_system_memory, state_component_assign, state_component_update in Hassign.
  destruct Hassign with Guest_OSs as [(Heq, Hinterp) | (Hnoteq, Hinterp)]; try discriminate.
  unfold state_component_interp in Hinterp; rewrite <- Hinterp; auto.

  (* New_untrusted *)
  unfold pre_new_untrusted in H_pre.
  destruct H_pre as (Huntrusted, (ma, hypma, (Hpt, (Hmaown, (Hacc, (Hhcall, Hyper)))))).
  unfold get_os_hypercall in Hhcall.
  destruct Hhcall as (os', Hval, Hcall).
  destruct no_hcall_s with o as (os'', Hvalos'', hcall_os'').
  assert (os' = os'') as eqos.
  rewrite <- Hval in Hvalos''; injection  Hvalos''; auto.
  rewrite eqos in Hcall; rewrite Hcall in hcall_os''; discriminate.

  (* New_trusted *)  
  unfold post_new_trusted, post_all_new in H_post.
  destruct H_post as (ma, Hhyp, (pta, osc, (Hnewmen, Hupd, (os', Hval_os', Hassign)))).
  unfold state_components_assign, state_components_update in Hassign.
  destruct Hassign with Guest_OSs as [((Heq1, Heq2), Hinterp) | (Hnoteq, Hinterp)]; try discriminate.
  elim Heq2; auto.
  unfold state_component_interp in Hinterp.
  destruct Hinterp as (Hmem, eqs'); rewrite eqs'; simpl.
  intro osi; destruct (no_hcall_s osi) as (os'', Hvalos'', Hhcall).
  elim (os_ident_eq (st_active_os s) osi); intro eq_aos_osi.
  rewrite eq_aos_osi; rewrite lemma_map_add_apply; auto.
  exists (OS None (os_current_page os')); auto.
  exists os''. rewrite Hvalos''.
  symmetry; rewrite lemma_map_add_apply_distinct; auto.
  auto.

  (* New_hyper *)
  unfold post_new_hyper, post_all_new in H_post.
  destruct H_post as (aos, Heq, (pt_addr, H_os_current, (Hnewmen, H_update, Hassign))). 
  unfold state_component_assign, state_component_update in Hassign.
  unfold IF_then_else in Hassign.
  destruct Hassign with Guest_OSs as [(Heqcomp1, Hinterp) | (Heqcomp2, Hinterp)]; try discriminate.
  unfold state_component_interp in Hinterp; rewrite <- Hinterp; auto.

  (* Del_untrusted *)
  unfold pre_del_untrusted in H_pre.
  destruct H_pre as (Huntrusted, (Hhcall, Hrest)).
  unfold get_os_hypercall in Hhcall.
  destruct Hhcall as (os, Hva, Hos_hcall).
  destruct no_hcall_s with o as (os', Hval_o, os'_not_hcall).
  rewrite <- Hva in Hval_o.
  simplify_eq Hval_o; intro eq_os'_os; rewrite eq_os'_os in os'_not_hcall.
  rewrite os'_not_hcall in Hos_hcall; discriminate.

  (* Del_trusted *)
  unfold post_del_trusted, post_all_del in H_post.
  destruct H_post as (Hpta, Hcurr, (Hnewmen, Hdrop, Hassign)).
  unfold assign_system_memory, state_component_assign, state_component_update in Hassign.
  destruct Hassign with Guest_OSs as [(Heqcomp1, Hinterp) | (Heqcomp2, Hinterp)]; try discriminate.
  unfold state_component_interp in Hinterp; rewrite <- Hinterp; auto.

  (* Del_hyper *)
  unfold post_del_hyper, post_all_del in H_post.
  destruct H_post as (Hpta, Hcurr, (Hnewmen, Hdrop, Hassign)). 
  unfold assign_system_memory, state_component_assign, state_component_update in Hassign.
  destruct Hassign with Guest_OSs as [(Heqcomp1, Hinterp) | (Heqcomp2, Hinterp)]; try discriminate.
  unfold state_component_interp in Hinterp; rewrite <- Hinterp; auto.

  (* Switch *)
  unfold post_switch in H_post.
  unfold state_component_assign, state_component_update in H_post.
  destruct H_post with Guest_OSs as [(Heqcomp1, Hinterp) | (Heqcomp2, Hinterp)]; try discriminate.
  unfold state_component_interp in Hinterp; rewrite <- Hinterp; auto.

  (* Lswitch_trusted *)
  unfold post_lswitch_trusted, post_all_lswitch in H_post.
  unfold state_component_assign, state_component_update in H_post.
  destruct H_post as (os', Hval, Hassign). 
  destruct Hassign with Guest_OSs as [(Heqcomp1, Hinterp) | (Heqcomp2, Hinterp)].
  unfold state_component_interp in Hinterp; rewrite Hinterp.
  intro osi; destruct no_hcall_s with osi as (os'', Hval', os''_not_hcall).
  elim (os_ident_eq (st_active_os s) osi); intro eq_aos_osi.
  rewrite eq_aos_osi; rewrite lemma_map_add_apply; auto.
  exists (OS None p); auto.
  exists os''. rewrite Hval'.
  symmetry; rewrite lemma_map_add_apply_distinct; auto.
  auto.
  elim Heqcomp2; auto.

  (* Lswitch_untrusted *)
  unfold pre_lswitch_untrusted in H_pre.
  destruct H_pre as (Huntrusted, (os', Hval, (Hhcall, Hrest))).
  destruct no_hcall_s with o as (os'', Hval_o, os'_not_hcall).
  rewrite <- Hval in Hval_o.
  simplify_eq Hval_o; intro eq_os'_os; rewrite eq_os'_os in os'_not_hcall.
  rewrite os'_not_hcall in Hhcall; discriminate.
  
  (* Hcall *)
  elim (a_not_Hcall h); auto.
  
  (* Ret_ctrl *)
  unfold post_ret_ctrl in H_post.
  unfold state_components_assign, state_components_update  in H_post.
  destruct H_post with Guest_OSs as [(Heqcomp1 , Hinterp) | (Heqcomp2, Hinterp)].
  unfold state_component_interp in Hinterp; rewrite <- Hinterp; auto.
  elim Heqcomp2; split; discriminate.

  (* Chmod *)
  unfold post_chmod in H_post.
  unfold state_components_assign, state_components_update  in H_post.
  destruct H_post as [(Htrusted, Hassign) | (Huntrusted, Hassign)].
  destruct Hassign with Guest_OSs as [(Heqcomp1 , Hinterp) | (Heqcomp2, Hinterp)].
  unfold state_component_interp in Hinterp; rewrite <- Hinterp; auto.
  elim Heqcomp2; split; discriminate.
  destruct Hassign with Guest_OSs as [(Heqcomp1 , Hinterp) | (Heqcomp2, Hinterp)].
  unfold state_component_interp in Hinterp; rewrite <- Hinterp; auto.
  elim Heqcomp2; split; discriminate.

  (* Pin_trusted *)
  unfold post_page_pin_trusted, post_all_pin in H_post.
  destruct H_post as (ma, (pg, val, (pgown, Hother))).
  destruct Hother as (pgcon, (p2m, hypp2m, (os', valos', eqs'))).
  rewrite eqs'; clear eqs'; simpl.
  intro osi; destruct no_hcall_s with osi as (os'', Hval', os''_not_hcall).
  elim (os_ident_eq (st_active_os s) osi); intro eq_aos_osi.
  rewrite eq_aos_osi; rewrite lemma_map_add_apply; auto.
  exists (OS None (os_current_page os')); auto.
  exists os''. rewrite Hval'.
  symmetry; rewrite lemma_map_add_apply_distinct; auto.
  auto.

  (* Unpin_trusted *)
  unfold post_page_unpin_trusted, post_all_unpin in H_post.
  destruct H_post as (p2m, hypp2m, (ma', valma', (os', valos', eqs'))).
  rewrite eqs'; clear eqs'; simpl.
  intro osi; destruct no_hcall_s with osi as (os'', Hval', os''_not_hcall).
  elim (os_ident_eq (st_active_os s) osi); intro eq_aos_osi.
  rewrite eq_aos_osi; rewrite lemma_map_add_apply; auto.
  exists (OS None (os_current_page os')); auto.
  exists os''. rewrite Hval'.
  symmetry; rewrite lemma_map_add_apply_distinct; auto.
  auto.
  
  (* Pin_untrusted *)
  unfold pre_page_pin_untrusted in H_pre.
  destruct H_pre as (Huntrusted, (Hhcall, Hrest)).
  unfold get_os_hypercall in Hhcall.
  destruct Hhcall as (os, Hva, Hos_hcall).
  destruct no_hcall_s with o as (os', Hval_o, os'_not_hcall).
  rewrite <- Hva in Hval_o.
  simplify_eq Hval_o; intro eq_os'_os; rewrite eq_os'_os in os'_not_hcall.
  rewrite os'_not_hcall in Hos_hcall; discriminate.

  (* Unpin_untrusted *)
  unfold pre_page_unpin_untrusted in H_pre.
  destruct H_pre as (Huntrusted, (Hhcall, Hrest)).
  unfold get_os_hypercall in Hhcall.
  destruct Hhcall as (os, Hva, Hos_hcall).
  destruct no_hcall_s with o as (os', Hval_o, os'_not_hcall).
  rewrite <- Hva in Hval_o.
  simplify_eq Hval_o; intro eq_os'_os; rewrite eq_os'_os in os'_not_hcall.
  rewrite os'_not_hcall in Hos_hcall; discriminate.
Qed.


Lemma running: forall (s:state),
  os_running s -> ~(hyper_running s).
Proof.
  unfold hyper_running, os_running, not; intros.
  rewrite H in H0; discriminate.
Qed.


(* osRunningNoHcall is an one-step invariant, for all action a<>Chmod *) 
Lemma OS_runnig_noHypercallState_inv : 
  forall (s s':state) (a: action),
  a <> Chmod ->
  osRunningNoHcall s ->
  valid_exec ctxt s a s' ->
  osRunningNoHcall s'.
Proof.
  intros s s' a a_not_Chmod.
  unfold osRunningNoHcall; intros osRunningNoHcall_s valid_exec_a .
  generalize (noHcall_inv s s' a); intro noHcall_inv_s_s'.
  induction a; inversion valid_exec_a as [s_vexec s'_vexec act_vexec H_pre H_post]; unfold pre_action in H_pre; unfold post_action in H_post.

  (* Silent *)
  unfold post_silent, same_state in H_post.
  rewrite <- H_post; auto.

  (* Read *)
  unfold post_read, post_all_read, same_state in H_post.
  rewrite <- H_post; auto.

  (* Read_hyper *)
  unfold post_read_hyper, post_all_read, same_state in H_post.
  rewrite <- H_post; auto.

  (* Write *)
  unfold pre_write in H_pre.
  destruct H_pre as (os_running, Hrest). 
  generalize (osRunningNoHcall_s (running s os_running)); intros.
  apply noHcall_inv_s_s'; [unfold not; intros; discriminate | auto | auto].

  (* Write_hyper *)
  unfold post_write_hyper, post_all_write in H_post.
  destruct H_post as (ma, va_ma, (newm, Hupd, Hassign)).
  unfold assign_system_memory, state_component_assign, state_component_update in Hassign.
  destruct Hassign with OS_Activity as [(Heq, Hinterp) | (Hnoteq, Hinterp)]; try discriminate.
  unfold state_component_interp in Hinterp; unfold noHcall.
  intro not_hyperr_s'. 
  unfold pre_write_hyper in H_pre.
  destruct H_pre as (hyperr, Hrest).
  unfold hyper_running in hyperr, not_hyperr_s'.
  rewrite hyperr in Hinterp; rewrite <- Hinterp in not_hyperr_s'.
  elim not_hyperr_s'; auto.

  (* New_untrusted *)
  unfold pre_new_untrusted in H_pre.
  destruct H_pre as (Huntrusted, (ma, hypma, (Hpt, (Hmaown, (Hacc, (Hhcall, hyperr)))))).
  intro not_hyperr_s'.
  unfold post_new_untrusted, post_all_new in H_post.
  destruct H_post as (ma', hypma', (ptadr, oscurr, (newmen', Hupd, (os', Hval, Hassign)))).
  unfold state_components_assign, state_components_update in Hassign.
  destruct Hassign with OS_Activity as [((eq1, eq2), Hinterp) | (Hnoteq, Hinterp)]; try discriminate.
  unfold state_component_interp in Hinterp.
  unfold hyper_running in hyperr, not_hyperr_s'.
  rewrite hyperr in Hinterp; rewrite <- Hinterp in not_hyperr_s'.
  elim not_hyperr_s'; auto.
  elim Hnoteq; split; discriminate.

  (* New_trusted *)  
  unfold pre_new_trusted in H_pre.
  destruct H_pre as (ma, Hhyp, (pgt, (Howner, (Hacc, (Htrusted, os_running))))).
  generalize (osRunningNoHcall_s (running s os_running)); intros.
  apply noHcall_inv_s_s'; [unfold not; intros; discriminate | auto | auto].

  (* New_hyper *)
  unfold pre_new_hyper in H_pre.
  destruct H_pre as (hyperr, Hrest).
  intro not_hyperr_s'.
  unfold post_new_hyper, post_all_new in H_post.
  destruct H_post as (aos, Heq, (pt_addr, H_os_current, (Hnewmen, H_update, Hassign))). 
  unfold state_component_assign, state_component_update in Hassign.
  unfold IF_then_else in Hassign.
  destruct Hassign with OS_Activity as [(Heqcomp1, Hinterp) | (Heqcomp2, Hinterp)]; try discriminate.
  unfold state_component_interp in Hinterp.
  unfold hyper_running in hyperr, not_hyperr_s'.
  rewrite hyperr in Hinterp; rewrite <- Hinterp in not_hyperr_s'.
  elim not_hyperr_s'; auto.

  (* Del_untrusted *)
  unfold pre_del_untrusted in H_pre.
  destruct H_pre as (Huntrusted, (Hhcall, (os', Hval', (Hp2m, Hhyp, (ptadr, Hval'', (Hpre_all, (hyperr, Hacc))))))).
  intro not_hyperr_s'.
  unfold post_del_untrusted in H_post.
  destruct H_post as (os'', Hval''', (Hp2m', Hhyp', (ptadr', Hval4, (Hnewmen, Hdrop, Hassign)))).
  unfold state_components_assign, state_components_update in Hassign.
  destruct Hassign with OS_Activity as [((Heq1, Heq2), Hinterp) | (Heqcomp2, Hinterp)]; try discriminate.
  unfold state_component_interp in Hinterp.
  unfold hyper_running in hyperr, not_hyperr_s'.
  rewrite hyperr in Hinterp; rewrite <- Hinterp in not_hyperr_s'.
  elim not_hyperr_s'; auto.
  elim Heqcomp2; split; discriminate.

  (* Del_trusted *)
  unfold pre_del_trusted in H_pre.
  destruct H_pre as (ptaddr, Hcurr, (Hpre_all, (os_running, Hrest))). 
  generalize (osRunningNoHcall_s (running s os_running)); intros.
  apply noHcall_inv_s_s'; [unfold not; intros; discriminate | auto | auto].

  (* Del_hyper *)
  unfold pre_del_hyper in H_pre.
  destruct H_pre as (ptaddr, Hcurr, (Hpre_all, (os_running, Hrest))). 
  intro not_hyperr_s'.  
  unfold post_del_hyper, post_all_del in H_post.
  destruct H_post as (Hpta, Hcurr', (Hnewmen, Hdrop, Hassign)). 
  unfold assign_system_memory, state_component_assign, state_component_update in Hassign.
  destruct Hassign with OS_Activity as [(Heqcomp1, Hinterp) | (Heqcomp2, Hinterp)]; try discriminate.
  unfold state_component_interp in Hinterp.
  unfold hyper_running in os_running, not_hyperr_s'.
  rewrite os_running in Hinterp; rewrite <- Hinterp in not_hyperr_s'.
  elim not_hyperr_s'; auto.

  (* Switch *)
  unfold pre_switch in H_pre.
  destruct H_pre as (hyperr, Hrest). 
  intro not_hyperr_s'.  
  unfold post_switch in H_post.
  unfold state_component_assign, state_component_update in H_post.
  destruct H_post with OS_Activity as [(Heqcomp1, Hinterp) | (Heqcomp2, Hinterp)]; try discriminate.
  unfold state_component_interp in Hinterp.
  unfold hyper_running in hyperr, not_hyperr_s'.
  rewrite hyperr in Hinterp; rewrite <- Hinterp in not_hyperr_s'.
  elim not_hyperr_s'; auto.

  (* Lswitch_trusted *)
  unfold pre_lswitch_trusted in H_pre.
  destruct H_pre as (os_running, Hrest). 
  generalize (osRunningNoHcall_s (running s os_running)); intros.
  apply noHcall_inv_s_s'; [unfold not; intros; discriminate | auto | auto].

  (* Lswitch_untrusted *)
  unfold pre_lswitch_untrusted in H_pre.
  destruct H_pre as (Huntrusted, (os', Hval, (Hhcall, (hyperr, Hrest)))).
  intro not_hyperr_s'.  
  unfold post_lswitch_untrusted, post_all_lswitch in H_post.
  destruct H_post as (os'', Hval', Hassign). 
  unfold state_component_assign, state_component_update in Hassign.
  destruct Hassign with OS_Activity as [(Heqcomp1, Hinterp) | (Heqcomp2, Hinterp)]; try discriminate.
  unfold state_component_interp in Hinterp.
  unfold hyper_running in hyperr, not_hyperr_s'.
  rewrite hyperr in Hinterp; rewrite <- Hinterp in not_hyperr_s'.
  elim not_hyperr_s'; auto.
  
  (* Hcall *)
  unfold post_hcall in H_post.
  destruct H_post as (os', H_val, eqs').
  intro not_hyperr_s'.  
  unfold hyper_running in not_hyperr_s'.
  cut ((st_aos_activity s')=OSA_waiting); 
  [intro st_aos_activity_s'; rewrite st_aos_activity_s' in not_hyperr_s'; elim not_hyperr_s'; auto | rewrite eqs'; simpl; auto].
  
  (* Ret_ctrl *)
  unfold post_ret_ctrl in H_post.
  intro not_hyperr_s'.  
  unfold state_components_assign, state_components_update in H_post.
  destruct H_post with OS_Activity as [((Heq1, Heq2), Hinterp) | (Heqcomp2, Hinterp)]; try discriminate.
  unfold state_component_interp in Hinterp.
  unfold hyper_running in not_hyperr_s'.
  elim Heq2; auto.
  unfold state_component_interp in Hinterp; destruct Hinterp as (Hexcm, Hact).
  unfold hyper_running in not_hyperr_s'.
  rewrite Hact in not_hyperr_s'; elim not_hyperr_s'; auto.

  (* Chmod *)
  elim a_not_Chmod; auto.

  (* Pin_trusted *)
  unfold pre_page_pin_trusted in H_pre.
  destruct H_pre as (os_running, Hrest). 
  generalize (osRunningNoHcall_s (running s os_running)); intros.
  apply noHcall_inv_s_s'; [unfold not; intros; discriminate | auto | auto].

  (* Unpin_trusted *)
  unfold pre_page_unpin_trusted in H_pre.
  destruct H_pre as (os_running, Hrest). 
  generalize (osRunningNoHcall_s (running s os_running)); intros.
  apply noHcall_inv_s_s'; [unfold not; intros; discriminate | auto | auto].
  
  (* Pin_untrusted *)
  unfold pre_page_pin_untrusted in H_pre.
  destruct H_pre as (Huntrusted, (Hhcall, (hyperr, Hrest))).
  intro not_hyperr_s'.  
  unfold post_page_pin_untrusted, post_all_pin in H_post.
  destruct H_post as (ma, (pg, Hval, (Hown, (Hpage, (Hp2m, Hhyp, (os', Hval', eqs')))))). 
  unfold hyper_running in hyperr, not_hyperr_s'.
  cut ((st_aos_activity s')=(st_aos_activity s)); 
  [intro st_aos_activity_s'; rewrite st_aos_activity_s' in not_hyperr_s'; elim not_hyperr_s'; auto | auto].
  rewrite eqs'; simpl; auto.

  (* Unpin_untrusted *)
  unfold post_page_unpin_untrusted, post_all_unpin in H_post.
  destruct H_post as (Hp2m, Hhyp, (ma', Hval, (os', Hval', eqs'))). 
  intro not_hyperr_s'.  
  unfold pre_page_unpin_untrusted in H_pre.
  destruct H_pre as (Huntrusted, (Hhcall, (hyperr, Hrest))).
  unfold hyper_running in hyperr, not_hyperr_s'.
  cut ((st_aos_activity s')=(st_aos_activity s)); 
  [intro st_aos_activity_s'; rewrite st_aos_activity_s' in not_hyperr_s'; elim not_hyperr_s'; auto | auto].
  rewrite eqs'; simpl; auto.
Qed.

End Prop_hcalls.


Require Export Traces.
Require Export Exec.
Require Export Streams.

Section Prop_hcall_traces.

Variable ctxt : context.

(* A  co-inductive definition of trace where the Chmod action requires no hcall is pending *) 
CoInductive trace_imp : state_stream-> Prop :=
  Trace_imp_chmod : forall (s:exec_state)(str:state_stream),
           trace_imp str ->
           (ex_action s) = Chmod ->
           (* no hcall is pending *)  
	   noHcall (ex_state s) ->
	   system_transition ctxt s (hd str) ->
           trace_imp (Cons s str)
  | Trace_imp_not_chmod : forall (s:exec_state) (str:state_stream),
           trace_imp str ->
           (ex_action s) <> Chmod ->
	   system_transition ctxt s (hd str) ->
           trace_imp (Cons s str).


(* trace_imp <= trace *) 
Theorem incl_trace_imp_in_trace : 
  forall s: state_stream, trace_imp s -> trace ctxt s.
Proof.
(* proof by coinduction *)
  cofix str.
  intros str0 trace_imp_sstr0.
  destruct str0.
  inversion trace_imp_sstr0; constructor; auto.
Qed.


(* Condition about an initial state of an implementation of the system (for trace_imp) *)
Definition init_state_imp : exec_state -> Prop :=
  fun execst => osRunningNoHcall (ex_state execst).

(* run_imp: a trace starting from an initial state without hypercalls, if an OS is running *)             
Definition run_imp (str:state_stream): Prop := 
  init_state_imp (hd str) /\ trace_imp str.


(* Auxiliar definitions *)

(* the hypervisor is not running in (hd str)*)
Definition state_stream_imp_not_hyper_running (str:state_stream) := 
  ~ hyper_running (ex_state (hd str)).

Definition exist_state_stream_imp_not_hyper_running (str:state_stream) := 
  Exists state_stream_imp_not_hyper_running str.

(* if the hypervisor is not running, then no hypercall is pending in (hd str) *)
Definition state_stream_imp_osRunningNoHcall (str:state_stream) := 
  osRunningNoHcall (ex_state (hd str)).

(* no hypercall is pending in (hd str) *)
Definition state_stream_imp_noHcall (str:state_stream) := 
  noHcall (ex_state (hd str)).

Definition exist_state_stream_imp_noHcall (str:state_stream) := 
  Exists state_stream_imp_noHcall str.


(* osRunningNoHcall is an invariant for trace_imp *)
Lemma osRunningNoHcall_inv : 
             forall str, run_imp str -> 
             ForAll state_stream_imp_osRunningNoHcall str.
Proof.
(* proof by coinduction *)
  cofix str.
  intros str0 run_str0.
  destruct str0.
  inversion run_str0 as [Hinit Htrace_cons].
  simpl in Hinit.
  unfold init_state_imp in Hinit.
  inversion Htrace_cons as [es st Htrace_tl Hact noHcall_hd Hvalid_exec | es st Htrace_tl Hact Hvalid_exec].
  (* case: Chmod action *)
  unfold system_transition in Hvalid_exec.
  rewrite Hact in Hvalid_exec.
  (* using auxiliar lemma osRunningNoHcall_Chmod *)
  generalize (osRunningNoHcall_Chmod ctxt (ex_state e)(ex_state (hd str0)) noHcall_hd Hvalid_exec).
  intro H_osRunningNoHcall_Chmod.
  constructor.
  unfold state_stream_imp_osRunningNoHcall; simpl; auto.
  (* apply coinduction hypothesis *)
  simpl; apply str.
  unfold run_imp, init_state_imp; split.
  assumption.
  inversion Htrace_cons; auto.
  (* case: action <> Chomd *)
  unfold system_transition in Hvalid_exec.
  (* using auxiliar lemma OS_runnig_noHypercallState_inv *)
  generalize (OS_runnig_noHypercallState_inv ctxt (ex_state e) (ex_state (hd str0)) (ex_action e) Hact Hinit Hvalid_exec).
  intro H_osRunningNoHcall_hds.
  constructor.
  unfold state_stream_imp_osRunningNoHcall; simpl.
  assumption.
  (* apply coinduction hypothesis *)
  simpl; apply str.
  unfold run_imp, init_state_imp; split.
  assumption.
  inversion Htrace_cons; auto.
Qed.


(* If every execution of the hypervisor is finite, then the hypervisor calls terminate (every hypercall is executed) *)
Theorem hcalls_are_executed : 
             forall str, run_imp str -> 
             ForAll (fun str' => exist_state_stream_imp_not_hyper_running str') str ->
             ForAll (fun str' => exist_state_stream_imp_noHcall str') str.
Proof.
(* proof by coinduction *)
  cofix str.
  intros str0 run_str0 H_str0.
  destruct str0.
  constructor.
  (* proof for the first state *)
  inversion run_str0 as [Hinit Htrace_cons].
  (* using auxiliar lemma osRunningNoHcall_inv *)
  generalize (osRunningNoHcall_inv (Cons e str0) run_str0); intro H_osRunningNoHcall_inv.
  inversion H_str0 as [Hexists_hd Hexists_tl].
  (* induction in exist_state_stream_imp_not_hyper_running *)
  induction Hexists_hd as [str1 | str1 Exists].
  destruct str1.
  inversion H_osRunningNoHcall_inv as [H_osRunningNoHcall_e0 H_osRunningNoHcall_tl].
  unfold state_stream_imp_osRunningNoHcall, osRunningNoHcall  in H_osRunningNoHcall_e0.
  apply Here; unfold state_stream_imp_noHcall; auto.
  (* inductive case for exist_state_stream_imp_not_hyper_running *)
  apply Further.
  apply IHExists; auto. 
  inversion run_str0 as [H_init_hd_str1]; auto.
  unfold run_imp; split.
  unfold init_state_imp.
  inversion H_osRunningNoHcall_inv as [H_osRunningNoHcall_hd H_osRunningNoHcall_tl_str1]. 
  inversion H_osRunningNoHcall_tl_str1 as [H_osRunningNoHcall_hd_tl H_osRunningNoHcall_tl_tl_str1].
  unfold state_stream_imp_osRunningNoHcall in H_osRunningNoHcall_hd_tl; auto.
  inversion H; auto.
  unfold init_state_imp.
  inversion H_osRunningNoHcall_inv as [H_osRunningNoHcall_str1 H_osRunningNoHcall_tl_str1].
  inversion H_osRunningNoHcall_tl_str1; auto.
  inversion Htrace_cons; auto.
  inversion H_osRunningNoHcall_inv; auto.
  inversion Hexists_tl; auto.
  (* apply coinduction hypothesis *)
  simpl; apply str; auto.
  unfold run_imp in  run_str0; simpl; elim run_str0; intros.
  unfold run_imp; split.
  unfold init_state_imp.
  (* using auxiliar lemma osRunningNoHcall_inv *)
  generalize (osRunningNoHcall_inv (Cons e str0) run_str0); intro osRunningNoHcall_cons_str0.
  inversion osRunningNoHcall_cons_str0 as [H_osRunningNoHcall_hd H_osRunningNoHcall_tl].
  inversion H_osRunningNoHcall_tl as [H_osRunningNoHcall_hd_tl H_osRunningNoHcall_tl_tl].
  simpl in H_osRunningNoHcall_tl_tl; auto.
  inversion H0; auto.
  inversion H_str0; auto.
Qed.


End Prop_hcall_traces.