(***********************************************************************
 * File: Maps.v
 * $Revision: 364 $
 * $Date: 2010-11-04 14:16:55 -0200 (Thu, 04 Nov 2010) $
 * $LastChangedBy: jdcampo $
 * ********************************************************************
 * 
 ************************************************************************)

Require Export Exc.
Require Import List.

Set Implicit Arguments.

(* This module describes a naive implementation of a dictionary (maps) using
   association lists. *)

Notation "'If' c1 'then' c2 'else' c3" :=
  match c1 with
  | left _ => c2
  | right _ => c3
  end (at level 200).

Section Mapping_Definition.

(* The type of index for accessing the map *)
Variable index : Set.
Variable index_eq : forall x y : index, {x = y} + {x <> y}.

(* The type of information associated to each index *)
Variable info : Set.

Record item : Set := 
 Item
    (* the index *)
    {item_index : index;
    (* the data associated to the index *)
    item_info : info 
   }.

(* Mappings as lists *)
Definition mapping : Set := list item.

(* Empty map  *)
Definition map_empty : mapping := nil.


(* Map add *)
(* Adding of a new entry in the map. 
   If the key already exists in the map, the info is redefined. *)
Fixpoint map_add (mp : mapping) (idx : index) 
  (inf : info) : mapping :=
  let newit := (Item idx inf) in 
  match mp with
  | nil =>  newit :: map_empty
  | it :: mp1 =>
      If index_eq (item_index it) idx
      then newit :: mp1 
      else it :: map_add mp1 idx inf
  end.

(* Map apply *)
Fixpoint map_apply (mp : mapping) 
  (idx : index) : exc info index :=
  match mp with
  | nil => Error info idx
  | it :: mp1 =>
      If index_eq idx (item_index it) 
      then Value index (item_info it)
      else map_apply mp1 idx
  end.

(* Map drop *)
Fixpoint map_drop (mp : mapping) 
  (idx : index) : mapping :=
  match mp with
  | nil => nil 
  | it :: mp1 =>
      If index_eq idx (item_index it) 
      then map_drop mp1 idx
      else it :: map_drop mp1 idx
  end.
 
(* Several map observers *)

Definition map_valid_index (mp : mapping) (idx : index) : Prop :=
  exists it : _, map_apply mp idx = Value index it.

Definition map_at_index (mp : mapping) 
  (idx : index) (inf : info) : Prop :=
  map_apply mp idx = Value index inf.

(* Map add Prop *)
Definition map_add_yields (mp : mapping) 
  (idx : index) (inf : info) (newmp : mapping) : Prop :=
  map_add mp idx inf = newmp.

(* Map size *)
Definition map_size (mp : mapping) : nat := length mp.

(* Lists the items in the map *)
Definition map2list (mp : mapping) : list item := mp.

End Mapping_Definition.
