(***********************************************************************
 * File: State.v
 * $Revision: 364 $
 * $Date: 2010-11-04 14:16:55 -0200 (Thu, 04 Nov 2010) $
 * $LastChangedBy: jdcampo $
 *******************************************************************
 * This file specifies what is meant by a State
 * 
 ******************************************************************)

Require Export Maps.
Require Export OS.

(** A table of OSes entries *)
Definition oss_map := mapping os_ident os.

(** The hypervisor is modelled as a mapping that for each OS associates
   a mapping from physical to machine addresses *)
Definition map_physical_machine_addr := mapping padd madd.
Definition hypervisor := mapping os_ident map_physical_machine_addr. 

(** Structure for the memory of the system.
   A mapping that associates to a machine address the corresponding page.
   It has to be established that a machine address belongs at most to one OS *) 

Inductive page_owner : Set :=
 | Hyp_PO   : page_owner
 | Os_PO    : os_ident -> page_owner
 | No_Owner : page_owner.

Record page : Set :=
  Page
    { (** content of a page *)
      page_content  : content;
      (** page owner *)
      page_owned_by : page_owner
    }.

(** The memory of the system *) 
Definition system_memory := mapping madd page.

(* Modifiy the value of a RW page *)
Definition update_rw_page
  (ma:madd) (val:value) (mem mem':system_memory) : Prop :=
  exists2 p, Value _ p = map_apply madd_eq mem ma &
             let newp := Page (RW_c (Some val)) (page_owned_by p) in
               mem' = map_add madd_eq mem ma newp.

(* Modifiy the mapping of a PT page.
   Add va->ma to the PT stored in address pt_addr. *)
Definition update_pt_page
  (pt_addr:madd) (va:vadd) (ma:madd) (mem mem':system_memory) : Prop :=
  exists2 p, Value _ p = map_apply madd_eq mem pt_addr &
  	     exists2 oldmap, PT_c oldmap = page_content p &
                             let newmap := map_add vadd_eq oldmap va ma in
                             let newp := Page (PT_c newmap) (page_owned_by p) in
                               mem' = map_add madd_eq mem pt_addr newp.

Definition drop_va_pt_page
  (pt_addr:madd) (va:vadd) (mem mem':system_memory) : Prop :=
  exists2 p, Value _ p = map_apply madd_eq mem pt_addr &
  	     exists2 oldmap, PT_c oldmap = page_content p &
                             let newmap := map_drop vadd_eq oldmap va in
                             let newp := Page (PT_c newmap) (page_owned_by p) in
                               mem' = map_add madd_eq mem pt_addr newp.

(** Page types : gives the type of a page from its address. *)
Definition page_type (smem:system_memory)(ma: madd) : option ptype :=
  let pg := map_apply madd_eq smem ma in
  match pg with
  | Error _ => None
  | Value pg =>
    let pagecnt := page_content pg in
    match pagecnt with
      | RW_c _ => Some RW
      | PT_c _ => Some PT
      | _ => None
    end end.

(** Definition of the State of the system **)

Record state : Set :=
  State
     { (** Active OS. *)
       st_active_os : os_ident; 
       (** OSs *)
       st_oss : oss_map; 
       (** Execution mode: ([usr]/[svc]). *)
       st_exec_mode : exec_mode;
       (** Activity of Active OS. When the Hypervisor is executing, 
           the corresponding Active OS is waiting *)
       st_aos_activity : os_activity; 
       (** hypervisor *)
       st_hypervisor : hypervisor;
       (* The memory *)	
       st_memory : system_memory
     }.
