(***********************************************************************
 * File: State_proof.v
 * $Revision: 482 $
 * $Date: 2011-03-05 09:41:24 -0200 (Sat, 05 Mar 2011) $
 * $LastChangedBy: jdcampo $
 ***********************************************************************
 * Proofs of invariance of state components
 * 
 **********************************************************************)

Require Import Maps_proof.
Require Export State_valid.
Require Export Env.


(** Some useful tactics  *)

(* Performs a case elimination on e introducing the corresponding 
   equation e=<pattern> as an hypothesis *)

Ltac Case_Equality e :=
  cut (e = e); [ pattern e at 2 in |- *; case e | reflexivity ].

Ltac Case_Rewrite e eq :=
  cut (e = e);
   [ pattern e at 2 in |- *; case e; intro eq; rewrite eq | reflexivity ].


(* Proves the simple invariants of a validity property where only one
   component of the state is modified *)
Ltac Valid_EqComp_Inv :=
 intros s s' H_prop H_cmpeq; rewrite <- H_cmpeq; assumption.

(* Idem but with two components *)
Ltac Valid_EqComp2_Inv :=
intros s s' H_prop H_cmpeq H_cmpeq'; 
  rewrite <- H_cmpeq; rewrite <- H_cmpeq'; assumption.

(* Idem but with three components *)
Ltac Valid_EqComp3_Inv :=
intros s s' H_prop H_cmpeq H_cmpeq' H_cmpeq'';
 rewrite <- H_cmpeq;rewrite <- H_cmpeq';rewrite <- H_cmpeq'';assumption.

(* Splitted introduction of the propositions that specify a valid state *)
Ltac Intros_Valid_Hyp:=
intros (H_trstos_nothypcall,
       (H_running_nothypcall,
       (H_valid_exmd,
       (H_valid_osexmd,
       (H_valid_untosexmd,
       (H_valid_hyper,
       (H_valid_virtmap,
       (H_valid_currpg,
       (H_injective_hyper,
        H_va_has_pa))))))))).

Section Valid_EqComp_Invariants.
Variable ctxt : context.

(* The memory is an injective mapping *)
Lemma memory_injective_inv:
 forall (s s':state),
 injective_mapping madd_eq (st_memory s) ->
 st_memory s = st_memory s' ->
 injective_mapping madd_eq (st_memory s').
Proof.
Valid_EqComp_Inv.
Qed.

(* Trusted OSs do not make hypercalls *)

Lemma trusted_os_not_hypercall_inv:
  forall (s s':state),
  trusted_os_not_hypercall (ctxt_oss ctxt) (st_oss s) ->
  st_oss s = st_oss s' ->
  trusted_os_not_hypercall (ctxt_oss ctxt) (st_oss s').
Proof.
Valid_EqComp_Inv.
Qed.  

(* If active OS is running, it should not have a hypercall. *)

 Lemma running_os_not_hypercall_inv:
  forall (s s':state),
  running_os_not_hypercall (st_active_os s) (st_aos_activity s) (st_oss s) ->
  st_active_os s = st_active_os s' ->
  st_oss s = st_oss s' ->
  st_aos_activity s = st_aos_activity s' ->
  running_os_not_hypercall (st_active_os s') (st_aos_activity s') (st_oss s').
Proof.
Valid_EqComp3_Inv.
Qed.

(* If hypervisor is running (aos is waiting) then execution mode is
   supervisor *)

Lemma valid_hyper_exec_mode_inv:
  forall (s s':state),
  valid_hyper_exec_mode (st_aos_activity s)(st_exec_mode s) ->
  st_aos_activity s = st_aos_activity s' ->
  st_exec_mode s = st_exec_mode s' ->
  valid_hyper_exec_mode (st_aos_activity s')(st_exec_mode s').
Proof.
Valid_EqComp2_Inv.
Qed.

(* If trusted OS is running, execution mode is supervisor *)

Lemma valid_trusted_os_exec_mode_inv:
 forall (s s':state),
 valid_trusted_os_exec_mode (st_active_os s) (st_exec_mode s) (ctxt_oss ctxt) ->
  st_active_os s = st_active_os s' ->
  st_exec_mode s = st_exec_mode s' ->
 valid_trusted_os_exec_mode (st_active_os s') (st_exec_mode s') (ctxt_oss ctxt).
Proof.
Valid_EqComp2_Inv.
Qed.

(* If untrusted OS is running, execution mode is user *)

Lemma valid_untrusted_os_exec_mode_inv:
 forall (s s':state),
 valid_untrusted_os_exec_mode 
       (st_active_os s) (st_aos_activity s)(st_exec_mode s) (ctxt_oss ctxt) ->
 st_active_os s = st_active_os s' ->
 st_aos_activity s = st_aos_activity s' ->
 st_exec_mode s = st_exec_mode s' ->
 valid_untrusted_os_exec_mode 
       (st_active_os s') (st_aos_activity s')(st_exec_mode s') (ctxt_oss ctxt).
Proof.
Valid_EqComp3_Inv.
Qed.


(* Hypervisor should map an OS padd to a madd owned by that OS. *)

Lemma valid_hypervisor_inv:
  forall (s s':state),
  valid_hypervisor (st_hypervisor s) (st_memory s) ->
  st_hypervisor s = st_hypervisor s' ->     
  st_memory s = st_memory s' ->     
  valid_hypervisor (st_hypervisor s') (st_memory s').
Proof.
Valid_EqComp2_Inv.
Qed.

Lemma valid_virtual_mapping_inv:
  forall (s s':state),
  valid_virtual_mapping (st_memory s) ctxt ->
  st_memory s = st_memory s' ->     
  valid_virtual_mapping (st_memory s') ctxt.
Proof.
Valid_EqComp_Inv.
Qed.

(* Current page table of any OS is of type PT, and owned by that OS. *)

Lemma valid_current_page_inv:
  forall (s s':state),
  valid_current_page (st_hypervisor s) (st_memory s) (st_oss s) ->
  st_hypervisor s = st_hypervisor s' ->     
  st_memory s = st_memory s' ->     
  st_oss s = st_oss s' ->     
  valid_current_page (st_hypervisor s') (st_memory s') (st_oss s').
Proof.
Valid_EqComp3_Inv.
Qed.

Lemma injective_hyper_mappings_inv:
  forall (s s':state),
  injective_hyper_mappings (st_hypervisor s) ->
  st_hypervisor s = st_hypervisor s' ->     
  injective_hyper_mappings (st_hypervisor s').
Proof.
Valid_EqComp_Inv.
Qed.

Lemma va_has_valid_pa_inv:
  forall (s s':state),
  va_has_valid_pa (st_memory s) (st_hypervisor s) (ctxt) ->
  st_memory s = st_memory s' ->
  st_hypervisor s = st_hypervisor s' ->
  va_has_valid_pa (st_memory s') (st_hypervisor s') (ctxt).
Proof.
Valid_EqComp2_Inv.
Qed.



End Valid_EqComp_Invariants.

Hint Resolve trusted_os_not_hypercall_inv: valid_simple.
Hint Resolve running_os_not_hypercall_inv: valid_simple.
Hint Resolve valid_trusted_os_exec_mode_inv: valid_simple.
Hint Resolve valid_hyper_exec_mode_inv: valid_simple.
Hint Resolve valid_untrusted_os_exec_mode_inv: valid_simple.
Hint Resolve valid_hypervisor_inv: valid_simple.
Hint Resolve valid_virtual_mapping_inv: valid_simple.
Hint Resolve valid_current_page_inv : valid_simple.
Hint Resolve injective_hyper_mappings_inv : valid_simple.
Hint Resolve va_has_valid_pa_inv : valid_simple.


(* If H_eqcomp is a hypothesis that specifies that the components cmp' of the 
 * state remains unchanged by an action then this tactic extends the proof 
 * context with the equality that establishes that the componet is the same 
 * in the pre and post states *)

Ltac Equal_State_Component H_eqcomp cmp cmp' H_eqcmp :=
assert (diff : cmp' <> cmp);
   [intro;discriminate | 
    generalize (H_eqcomp cmp' diff);simpl;intro H_eqcmp;clear diff].

(*
(* Tactics to prove in one step the invariance of a validity property *)
*)

(* The validity property depends on the component cmp' and the action affects
 * component cmp. eauto uses the invariant lemma for that property *)

Ltac Simp_Inv_OneComp H_eqcmp H_eqcmp' cmp cmp' :=
  Equal_State_Component H_eqcmp cmp cmp' H_eqcmp';
    eauto with valid_simple.

(* The property depends on two state components *)
Ltac Simp_Inv_TwoComps H_eqcmp H_eqcmp' H_eqcmp'' cmp cmp' cmp'':=
  Equal_State_Component H_eqcmp cmp cmp' H_eqcmp';
    Equal_State_Component H_eqcmp cmp cmp'' H_eqcmp'';
      eauto with valid_simple. 

(* The property depends on three state components *)
Ltac Simp_Inv_ThreeComps 
     H_eqcmp H_eqcmp' H_eqcmp'' H_eqcmp''' cmp cmp' cmp'' cmp''':=
  Equal_State_Component H_eqcmp cmp cmp' H_eqcmp';
    Equal_State_Component H_eqcmp cmp cmp'' H_eqcmp'';
      Equal_State_Component H_eqcmp cmp cmp''' H_eqcmp''';
        eauto with valid_simple. 


Lemma lemma_same_hypervisor_mapping :
  forall (osi:os_ident) (hyp:hypervisor) (p2m p2m':mapping padd madd),
    hypervisor_mapping osi hyp p2m ->
    hypervisor_mapping osi hyp p2m' ->
    p2m = p2m'.
Proof.
  unfold hypervisor_mapping.
  intros osi hyp p2m p2m' H_hyp_map H_hyp_map'.
  rewrite <- H_hyp_map' in H_hyp_map.
  injection H_hyp_map; auto.
Qed.
  
Lemma lemma_page_type_exists_page : forall (smem:system_memory) (ma:madd),
  page_type smem ma <> None ->
  exists pg:page, Value madd pg = map_apply madd_eq smem ma.
Proof.
  intros smem ma.
  unfold page_type.
  case (map_apply madd_eq smem ma).
  intros p H.
  exists p.
  reflexivity.
  intros m H_abs.
  contradiction H_abs; reflexivity.
Qed.

Lemma lemma_page_table_exists_map : forall (smem:system_memory) (pg:page) (pg_ma:madd),
    Value madd pg = map_apply madd_eq smem pg_ma ->
    page_type smem pg_ma=Some PT ->
    exists pt_map : mapping vadd madd, 
    PT_c pt_map = page_content pg.
Proof.
  intros smem pg pg_ma H_pg_ma.
  unfold page_type.
  rewrite <- H_pg_ma.

  case (page_content pg).
  intros H H_abs ;inversion H_abs.
  intros map H_triv.
  exists map.
  reflexivity.
  intro H_abs;inversion H_abs.
Qed.

Lemma hyp_mapped_same_maddr : forall (s:state) (osi:os_ident) (pa:padd) (ma1 ma2:madd),
  hypervisor_mapping_madd (st_hypervisor s) osi pa ma1 ->
  hypervisor_mapping_madd (st_hypervisor s) osi pa ma2 ->
  ma1 = ma2.
Proof.
  intros s osi pa ma1 ma2 H_ma1_map H_ma2_map.
  unfold hypervisor_mapping_madd in *.
  destruct H_ma1_map as [p2m1 H_p2m1_map H_ma1_map].
  destruct H_ma2_map as [p2m2 H_p2m2_map H_ma2_map].
  rewrite <- H_p2m1_map in H_p2m2_map.
  injection H_p2m2_map; intro; subst p2m2; clear H_p2m2_map.
  rewrite <- H_ma1_map in H_ma2_map.
  injection H_ma2_map; auto.
Qed.

Lemma hyp_mapped_same_paddr : forall (ctxt:context) (s:state) (osi:os_ident) (pa1 pa2:padd) (ma:madd),
  valid_state ctxt s ->
  hypervisor_mapping_madd (st_hypervisor s) osi pa1 ma ->
  hypervisor_mapping_madd (st_hypervisor s) osi pa2 ma ->
  pa1 = pa2.
Proof.
  intros ctxt s osi pa1 pa2 ma H_s_valid H_ma_map1 H_ma_map2.
  unfold hypervisor_mapping_madd in *.
  destruct H_ma_map1 as [p2m1 H_p2m1_map H_ma_map1].
  destruct H_ma_map2 as [p2m2 H_p2m2_map H_ma_map2].
  rewrite <- H_p2m1_map in H_p2m2_map.
  injection H_p2m2_map; intro; subst p2m2; clear H_p2m2_map.
  generalize H_s_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_valid_currpg.
  unfold injective_hyper_mappings in H_injective_hyper.
  rewrite H_injective_hyper with osi p2m1 pa1 pa2; auto.
  rewrite <- H_ma_map1.
  assumption.
Qed.