(***********************************************************************
 * File: Traces.v
 * $Revision: 454 $
 * $Date: 2011-01-21 09:51:03 -0200 (Fri, 21 Jan 2011) $
 * $LastChangedBy: jdcampo $
 ***********************************************************************
 * This module introduces platform execution traces 
 **********************************************************************)


Require Export Exec.
Require Export Exec_invariant.
Require Export Streams.

Section Traces.

Variable ctxt : context.

(* State and transition *)
Record exec_state : Set :=
  Ex_State
    {ex_state : state;
     ex_action  : action
    }.

(* We introduce a predicate that specifies what is meant by an initial state of the system.
   Right now we shall only require for it to be a valid state. 
   Later we shall surely enrich this definition
*)

Definition init_state : exec_state -> Prop :=
  fun execst => valid_state ctxt (ex_state execst).


(* This notion of transition makes it possible to predicate over the states and the action
   involved in a valid execution *)

Definition system_transition (src tgt : exec_state) : Prop :=
 
   let s := ex_state src in
   let s' := ex_state tgt in
   let a := ex_action src in
   valid_exec ctxt s a s'.    

Definition state_prop:= exec_state -> Prop.

(* Invariants *)

Definition leads_to 
  (P Q : state_prop) : Prop := 
   forall s s': exec_state,
    P s ->  
    system_transition s s' ->  
    Q s'.

Definition invariant: state_prop -> Prop := fun P => leads_to P P.

(* Streams of states, formulas and temporal operators *)

Definition state_stream := (Stream exec_state).

Definition stream_prop := state_stream->Prop.

Definition binary_state_prop := state -> state -> Prop.

Definition lift_unary (P:state_prop) : stream_prop := fun str => P (hd str).

Definition lift_binary (P:binary_state_prop) : stream_prop := 
    fun str => P (ex_state (hd str)) (ex_state (hd (tl str))).


(****** Traces and runs *********)

(* A trace is a stream of execution states (si,ai) for i=1.., s.t. si -> s(i+1),
   where -> is valid_exec  *)

(* A  co-inductive fdefinition of trace *) 
CoInductive trace : state_stream-> Prop :=
  Trace : forall (s:exec_state)(str:state_stream),
           trace str ->
	   system_transition s (hd str) ->
           trace (Cons s str).

(* run: a trace starting from an initial state *)             
Definition run (str:state_stream): Prop := 
  init_state (hd str) /\ trace str.

(* The following lemmas shall be extensively used when proving invariants 
   (safety properties) of platform runs *)

(* If a trace is a run, all of its sub-traces are also runs *)

Lemma run_inv : forall str, run str -> run (tl str).
Proof.
intros str; unfold run.
case str.
simpl; unfold init_state.
intros e s (H_valide,H_trace).
inversion H_trace.
split.
(* valid state hdtl *)
inversion H2.
rewrite <- H7. 
apply valid_state_invariant with s1 act.
rewrite H5. assumption.
constructor; rewrite H5; rewrite H6; try (rewrite H7); assumption.
(* trace s *)
assumption.
Qed.

(* Runs and valid state *)

Definition valid_state_stream (str:state_stream) := 
  valid_state ctxt (ex_state (hd str)).

Lemma state_of_run_is_valid: 
             forall str, run str -> valid_state_stream str.
Proof.
intros str; case str.
intros e s H_run; unfold valid_state_stream; simpl.
inversion H_run.
unfold init_state in H; simpl in H.
assumption.
Qed.

(* Invariant: All the states of a run are are valid states
   This lemma shall be useful to prove safety properties using
   the co-induction principle for ForAll, represented by the Coq term 
   ForAll_coind. The proof of this lemma itself illustrates the use of
  that principle
*)

Lemma states_of_run_are_valid : 
             forall str, run str -> ForAll valid_state_stream str.
Proof.
intros str H_run.
apply (ForAll_coind valid_state_stream run).
exact state_of_run_is_valid.
exact run_inv.
assumption.
Qed.

End Traces.

Section Safety.

Variable ctxt : context.


(***********************)
(******  Safety *********)
(***********************)

(* A Property is safe for a run of the system if it holds for all state of the execution trace *)

Definition safe (P:state -> state -> Prop) : Prop := 
  forall str:state_stream, run ctxt str -> ForAll (lift_binary P) str.

End Safety.