(***********************************************************************
 * File: Write_Isolation.v
 * $Revision: 399 $
 * $Date: 2010-12-26 17:56:13 -0200 (Sun, 26 Dec 2010) $
 * $LastChangedBy: cluna $
 ***********************************************************************
 * This files proves write isolation property on valid executions. 
 **********************************************************************)

Require Import State_valid.
Require Import Env.
Require Import Exec.
Require Import Exec_invariant.
Require Export Observers.

Section Write_Isolation.

Variable ctxt : context.

(** Lemmas for write_isolation *)

(* If states differ only in one page owned by the AOS,
   all pages owned by other OSs or by the hypervisor do not change. *) 
Lemma same_not_owned_page :
  forall (s s':state) (new_ma:madd) (old_pg new_pg:page),
    Value _ old_pg = map_apply madd_eq (st_memory s) new_ma -> 
    st_memory s' = map_add madd_eq (st_memory s) new_ma new_pg ->
    Os_PO (st_active_os s) = page_owned_by old_pg ->
    (forall (ma:madd) (pg:page), Value _ pg = map_apply madd_eq (st_memory s) ma ->
      ~ owner_of_page pg (st_active_os s) ->
      Value _ pg = map_apply madd_eq (st_memory s') ma).
Proof.
  intros s s' new_ma old_pg new_pg H_old_pg_map H_smem'_map H_old_pg_owner
    ma pg H_pg_map H_pg_owner.

  rewrite H_smem'_map.

  rewrite H_pg_map; symmetry.
  apply lemma_map_add_apply_distinct.

  intro H_ma_eq_new_ma.

  assert (pg = old_pg).
  
  subst new_ma.
  rewrite <- H_pg_map in H_old_pg_map.
  injection H_old_pg_map; auto.

  subst old_pg; eauto. 
Qed.


(* Executing a valid write does not change other memory. *)
Lemma lemma_write_action_isolation :
  forall (s s':state) (va_write:vadd) (val:value),
    valid_state ctxt s ->
    valid_exec ctxt s (Write va_write val) s' ->
    (forall (ma:madd) (pg:page), Value _ pg = map_apply madd_eq (st_memory s) ma ->
      ~ owner_of_page pg (st_active_os s) ->
      Value _ pg = map_apply madd_eq (st_memory s') ma).
Proof.
  intros s s' va_write val H_s_valid H_valid_exec ma pg H_pg_map H_pg_owner.

  inversion_clear H_valid_exec as [s_vexec s'_vexec act_vexec H_pre H_post].

  unfold post_action, post_write, post_all_write in H_post.
  destruct H_post as (ma_write, H_ma_write_map, (smem', H_up_rw, H_assign_smem')).

  unfold update_rw_page in H_up_rw.
  destruct H_up_rw as (pg_write, H_pg_write_map, H_smem'_map). (* pg_write is the page that
                                                                  the action changes *)

  (* Since this is a valid write, pg_write must be owned by the AOS *)
  assert (Os_PO (st_active_os s) = page_owned_by pg_write).
  unfold pre_action, pre_write in H_pre.
  destruct H_pre as (H_os_running, (H_os_accessible, (ma_write', H_ma_write'_map, H_ptype))).

  unfold va_mapped_to_ma in H_ma_write_map.
  destruct H_ma_write_map as (pt_addr, H_curr_pt, (mp, H_mp_map, H_ma_write_map)).

  generalize H_s_valid; clear H_s_valid; Intros_Valid_Hyp.
  unfold valid_virtual_mapping in H_valid_virtmap.

  unfold os_current_vadd_mapping_madd in H_curr_pt.
  destruct H_curr_pt as (aos, H_aos_map, (p2m, H_hyp_mapping, H_pt_addr_map)).
  unfold hypervisor_mapping in H_hyp_mapping.
  unfold vadd_mapping in H_mp_map.
  destruct H_mp_map as (pt_page, H_pt_page_map, H_pt_page_content).
  
  destruct H_valid_virtmap with (osi := st_active_os s) (pt_addr:=pt_addr) (pt:=pt_page)
    (mp:=mp) (va:=va_write) (ma:=ma_write) as [pg' H_pg'_map (H_owner, H_hyp)]; auto.

  unfold valid_hypervisor in H_valid_hyper.
  destruct H_valid_hyper with (osi:=(st_active_os s)) (pa:=(os_current_page aos)) (p2m:=p2m); auto.
  destruct H as (pt_addr', H_pt_addr'_map, H_pt_addr'_owner).
  assert (pt_addr'=pt_addr).
  rewrite <- H_pt_addr'_map in H_pt_addr_map; injection H_pt_addr_map; auto.
  subst pt_addr'; clear H_pt_addr'_map.
  unfold maddress_owner in H_pt_addr'_owner.
  rewrite <- H_pt_page_map in H_pt_addr'_owner.
  assumption.
  cut False. contradiction.
  rewrite H in H_pt_addr_map.
  discriminate H_pt_addr_map.
  
  (* pg' and pg_write are the pages in smem at address ma_write *)
  assert (pg_write = pg').
  rewrite <- H_pg'_map in H_pg_write_map; injection H_pg_write_map; auto.
  
  subst pg'.
  symmetry; apply H_owner; auto.

  (* Now that we know that pg_write is owned by the AOS, use the lemma *)
  apply same_not_owned_page with
      s ma_write pg_write (Page (RW_c (Some val)) (page_owned_by pg_write));
    auto.

  unfold assign_system_memory, state_component_assign,
    state_component_update in H_assign_smem'.
  destruct H_assign_smem' with (System_Memory) as [(H_trivial, H_same_mem) | (H_abs, H_state_assign')].
  
  unfold state_component_interp in H_same_mem; rewrite H_same_mem; auto.

  contradiction H_abs; reflexivity.
Qed.

(* Executing a valid new_trusted does not change other memory. *)
Lemma lemma_new_trusted_isolation :
  forall (s s':state) (new_va:vadd) (new_pa:padd),
    valid_state ctxt s ->
    valid_exec ctxt s (New_trusted new_va new_pa) s' ->
    (forall (ma:madd) (pg:page), Value _ pg = map_apply madd_eq (st_memory s) ma ->
    ~ owner_of_page pg (st_active_os s) ->      
    Value _ pg = map_apply madd_eq (st_memory s') ma).
Proof. 
  intros s s' new_va new_pa H_s_valid H_valid_exec
    ma pg H_pg_map H_pg_owner.

  inversion_clear H_valid_exec as [s_vexec s'_vexec act_vexec H_pre H_post].

  unfold post_action, post_new_trusted, post_all_new in H_post.
  destruct H_post as (new_ma, H_hyp_map_madd, (pt_addr, H_curr_map,
    (smem', H_up_pt_page, (aos, H_aos_map, H_s')))).

  unfold update_pt_page in H_up_pt_page.

  (* pt_page is the current PT of the AOS. *)
  destruct H_up_pt_page as (pt_page, H_pt_page_map, (oldmap, H_pt_page_content, H_smem')).

  
  apply same_not_owned_page with s pt_addr pt_page
    (Page (PT_c (map_add vadd_eq oldmap new_va new_ma)) (page_owned_by pt_page)); auto.

  (* We now need to prove that pt_page is the only page that changes in s'. *)
  unfold state_components_assign, state_components_update in H_s'.
  destruct H_s' with System_Memory as [((H_abs, H_triv), H_interp) |
    (H_triv, (H_smem'_interp, H_interp2))].
  contradiction H_abs; auto.
  simpl in H_smem'_interp.
  rewrite H_smem'_interp.
  assumption.

  (* Now, prove that AOS is the owner of pt_page, using valid_current_page *)
  generalize H_s_valid; Intros_Valid_Hyp.
  clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
    H_valid_untosexmd H_valid_hyper H_valid_virtmap.

  unfold os_current_vadd_mapping_madd in H_curr_map.
  destruct H_curr_map as (aos', H_aos'_map, (p2m, H_hyp_map, H_pt_addr_map)).

  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) aos p2m
    as (pt_addr', H_pt_addr'_map, (H_pt_addr_type, H_pt_addr_owner)); auto.

  (* Prove that objects in valid_current_page are the same we had *)
  assert (aos = aos').
  rewrite <- H_aos'_map in H_aos_map.
  injection H_aos_map; auto.
  subst aos'; clear H_aos'_map.

  assert (pt_addr' = pt_addr).
  rewrite <- H_pt_addr'_map in H_pt_addr_map.
  injection H_pt_addr_map; auto.
  subst pt_addr'; clear H_pt_addr'_map.

  (* Now we know the owner of pt_page *)
  rewrite <- H_pt_addr_owner.

  unfold maddress_owner.
  rewrite <- H_pt_page_map; auto.
Qed.
  
(* Executing a valid del_trusted does not change other memory. *)
Lemma lemma_del_trusted_isolation :
  forall (s s':state) (va:vadd),
    valid_state ctxt s ->
    valid_exec ctxt s (Del_trusted va) s' ->
    (forall (ma:madd) (pg:page), Value _ pg = map_apply madd_eq (st_memory s) ma ->
    ~ owner_of_page pg (st_active_os s) ->
     Value _ pg = map_apply madd_eq (st_memory s') ma).
Proof.
  intros s s' del_va H_s_valid H_valid_exec
    ma pg H_pg_map H_pg_owner.

  inversion_clear H_valid_exec as [s_vexec s'_vexec act_vexec H_pre H_post].

  unfold post_action, post_del_trusted, post_all_del in H_post.
  destruct H_post as (pt_addr, H_curr_map_madd, (smem', H_drop_pt_page, H_assign_smem)).

  unfold drop_va_pt_page in H_drop_pt_page.
  
  (* pt_page is the current PT of the AOS. *)
  destruct H_drop_pt_page as (pt_page, H_pt_page_map, (oldmap, H_pt_page_content, H_smem')).

  apply same_not_owned_page with s pt_addr pt_page
    (Page (PT_c (map_drop vadd_eq oldmap del_va)) (page_owned_by pt_page)); auto.

  (* We now need to prove that pt_page is the only page that changes in s'. *)
  unfold assign_system_memory, state_component_assign, state_component_update in H_assign_smem.
  destruct H_assign_smem with System_Memory as [ (H_triv, H_interp) |
    (H_abs, H_interp)].

  simpl in H_interp.
  rewrite H_interp.
  assumption.
  contradiction H_abs; auto.

  (* Now, prove that AOS is the owner of pt_page, using valid_current_page *)
  generalize H_s_valid; Intros_Valid_Hyp.
  clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
    H_valid_untosexmd H_valid_hyper H_valid_virtmap.

  unfold current_vadd_mapping_madd, os_current_vadd_mapping_madd in H_curr_map_madd.

  destruct H_curr_map_madd as (aos, H_aos_map, (p2m, H_hyp_map, H_pt_addr_map)).

  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) aos p2m
    as (pt_addr', H_pt_addr'_map, (H_pt_addr_type, H_pt_addr_owner)); auto.

  (* Prove that objects in valid_current_page are the same we had *)
  assert (pt_addr' = pt_addr).
  rewrite <- H_pt_addr'_map in H_pt_addr_map.
  injection H_pt_addr_map; auto.
  subst pt_addr'; clear H_pt_addr'_map.

  (* Now we know the owner of pt_page *)
  rewrite <- H_pt_addr_owner.

  unfold maddress_owner.
  rewrite <- H_pt_page_map; auto.
Qed.

(* Executing a valid lswitch (trusted or untrusted) does not change
   other memory. *)
Lemma lemma_lswitch_isolation :
  forall (s s':state) (osi:os_ident) (pa:padd) (a:action),
    valid_state ctxt s ->
    ((a = (Lswitch_trusted pa)) \/ (a = (Lswitch_untrusted osi pa))) -> 
    valid_exec ctxt s a s' ->
    (forall (ma:madd) (pg:page), Value _ pg = map_apply madd_eq (st_memory s) ma ->
    ~ owner_of_page pg (st_active_os s) ->
    Value _ pg = map_apply madd_eq (st_memory s') ma).
Proof.
  intros s s' osi pa a H_s_valid H_action H_valid_exec ma pg
    H_pg_map H_pg_owner.

  (* Memory doesn't change, so we only need to show
       (st_memory s' = st_memory s)
     for both actions. *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  destruct H_action as [H_lswitch_trusted | H_lswitch_untrusted]; subst a;
    unfold post_action, post_lswitch_trusted, post_lswitch_untrusted, post_all_lswitch in H_post;
      destruct H_post as (aos, H_aos_map, H_state_assign);
        unfold state_component_assign, state_component_update, state_component_interp in H_state_assign;
          destruct H_state_assign with System_Memory as [(H_abs, H_state_assign') | (H_trivial, H_same_mem)];
  try discriminate H_abs; rewrite <- H_same_mem; assumption.
Qed.  


(* Executing a valid page_pin_trusted does not change other memory used (reserved). *)
Lemma lemma_page_pin_trusted_isolation :
  forall (s s':state) (new_pa:padd) (page_type:ptype),
    valid_state ctxt s ->
    valid_exec ctxt s (Page_pin_trusted new_pa page_type) s' ->
    (forall (ma:madd) (pg:page), Value _ pg = map_apply madd_eq (st_memory s) ma ->
    ~ owner_of_page pg (st_active_os s) ->
    page_owned_by pg <> No_Owner ->
    Value _ pg = map_apply madd_eq (st_memory s') ma).
Proof.
  intros s s' new_pa page_type H_s_valid H_valid_exec ma pg H_pg_map H_pg_owner H_pg_no_owner.

  inversion_clear H_valid_exec as [s_vexec s'_vexec act_vexec H_pre H_post].

  unfold post_action, post_page_pin_trusted, post_all_pin in H_post.
  destruct H_post as (new_ma, (new_pg, H_new_pg_map, (H_new_pg_no_own,
      (H_new_pg_no_content, (p2m, H_hyp_map, (aos, H_aos_map, H_s')))))).

  (* Here we cannot use lemma same_not_owned_page, since the page that
     changes is not owned by AOS. We prove that new_ma (the madd of
     the page that is pinned), is different from any ma owned by other
     os, because it is owned by no-one. *)

  assert (new_ma <> ma).
  intros H_new_ma_eq_ma.
  subst new_ma.
  assert (new_pg = pg).
  rewrite <- H_new_pg_map in H_pg_map; injection H_pg_map; auto.

  subst new_pg.
  unfold owner_of_page in H_pg_owner.
  rewrite H_new_pg_no_own in H_pg_owner.
  rewrite H_new_pg_no_own in H_pg_no_owner.
  elim H_pg_no_owner; auto.

  (* Now we know that pg does not change, since the only memory change
     is in new_ma. *)
  replace (st_memory s') with (map_add madd_eq (st_memory s) new_ma
    (new_page (Some page_type) (Os_PO (st_active_os s)))).

  rewrite H_pg_map.
  symmetry; apply lemma_map_add_apply_distinct; auto.
  rewrite H_s'.
  simpl.
  reflexivity.
Qed.

(* Executing a valid page_unpin_trusted does not change other memory. *)
Lemma lemma_page_unpin_trusted_isolation :
  forall (s s':state) (pa:padd),
    valid_state ctxt s ->
    valid_exec ctxt s (Page_unpin_trusted pa) s' ->
    (forall (ma:madd) (pg:page), Value _ pg = map_apply madd_eq (st_memory s) ma ->
    ~ owner_of_page pg (st_active_os s) ->      
    Value _ pg = map_apply madd_eq (st_memory s') ma).
Proof.
  intros s s' unpin_pa H_s_valid H_valid_exec ma pg H_pg_map H_pg_owner.

  inversion_clear H_valid_exec as [s_vexec s'_vexec act_vexec H_pre H_post].

  unfold post_action, post_page_unpin_trusted, post_all_unpin in H_post.
  destruct H_post as (p2m, H_hyp_map, (unpin_ma, H_unpin_ma_map, (aos, H_aos_map, H_s'))).

  unfold pre_action, pre_page_unpin_trusted, pre_all_unpin in H_pre.
  destruct H_pre as (H_os_running, (H_trusted_aos, (aos', H_aos'_map,
    (H_pa_not_curr_pt, (p2m', H_hyp_map', (unpin_ma', H_unpin_ma'_map,
      (H_page_type, H_no_va_ma))))))).

  (* Prove that objects in pre and post are the same *)
  assert (p2m = p2m').
  apply lemma_same_hypervisor_mapping with (st_active_os s) (st_hypervisor s); auto.
  subst p2m'; clear H_hyp_map'.
  assert (unpin_ma = unpin_ma').
  rewrite <- H_unpin_ma'_map in H_unpin_ma_map.
  injection H_unpin_ma_map; auto.
  subst unpin_ma'; clear H_unpin_ma'_map.

  (* Obtain the changed page using lemma_page_type_exists_page *)
  assert (exists unpin_pg:page, Value madd unpin_pg = map_apply madd_eq (st_memory s) unpin_ma).
  apply lemma_page_type_exists_page; auto.
  destruct H as (unpin_pg, H_unpin_pg_map).

  (* Now we can apply lemma same_not_owned_page, using the changed page. *)
  apply same_not_owned_page with s unpin_ma unpin_pg (new_page None No_Owner); auto.

  (* Prove that page in unpin_ma is the only page that changes in s'. *)
  rewrite H_s'; simpl; reflexivity.

  (* Now, prove that AOS is the owner of unpin_pg, using valid_hypervisor. *)
  generalize H_s_valid; Intros_Valid_Hyp; clear H_trstos_nothypcall
    H_running_nothypcall H_valid_exmd H_valid_osexmd H_valid_untosexmd
    H_valid_virtmap H_valid_currpg.

  unfold valid_hypervisor in H_valid_hyper.
  destruct H_valid_hyper with (st_active_os s) unpin_pa p2m as
    [(unpin_ma', H_unpin_ma'_map, H_unpin_ma'_owner) | H_unpin_pa_error]; auto.

  (* valid_hypervisor is an or. Either the page is owned by AOS, or
     the page is not mapped. *)
  clear H_valid_hyper.
  
  assert (unpin_ma' = unpin_ma);
    [rewrite <- H_unpin_ma'_map in H_unpin_ma_map; injection H_unpin_ma_map; auto |
      subst unpin_ma'; clear H_unpin_ma'_map].

  unfold maddress_owner in H_unpin_ma'_owner.
  rewrite <- H_unpin_ma'_owner; rewrite <- H_unpin_pg_map; reflexivity.

  (* Absurd case where the page is not mapped in p2m.
     We know by H_unpin_ma_map that it is. *)
  cut False; [contradiction |
    rewrite H_unpin_pa_error in H_unpin_ma_map; discriminate].
Qed.

(** No action executed by an OS in a valid state changes memory that 
    does not belong to the AOS.
*)


(* Property write_isolation *)
Definition write_isolation' (s s':state) : Prop :=
 forall (a:action),
       valid_state ctxt s ->
       ~(hyper_running s) ->
       valid_exec ctxt s a s' ->     
       forall (ma:madd) (pg:page), 
          content_of_address s ma pg ->
          ~ owner_of_page pg (st_active_os s) ->
          page_owned_by pg <> No_Owner ->
          content_of_address s' ma pg.


Theorem write_isolation_proof' : forall (s s':state), write_isolation' s s'.
Proof.
  intros s s' a H_s_valid H_not_hyp_run H_valid_exec ma pg H_pg_map H_pg_owner H_pg_no_owner.

  (* Prove by cases on the action. Actions where the state does not
     change, or the hyper is running are proved directly. Others have
     their own lemma. *)
  induction a.

  (* Silent action (state does not change). *)
  inversion_clear H_valid_exec.
  unfold post_action, post_silent, same_state in H0.
  rewrite <- H0.
  assumption.

  (* Read action (state does not change) *)
  inversion_clear H_valid_exec.
  unfold post_action, post_read, post_read_hyper,
    post_all_read, same_state in H0.
  rewrite <- H0.
  assumption.

  (* Read hyper action (state does not change) *)
  inversion_clear H_valid_exec.
  unfold post_action, post_read, post_read_hyper,
    post_all_read, same_state in H0.
  rewrite <- H0.
  assumption.

  (* Write action *)
  apply lemma_write_action_isolation with s v v0; auto.
  
  (* Write Hyper (hyper is running). *)
  inversion_clear H_valid_exec.
  unfold pre_action, pre_write_hyper, hyper_running in H.
  unfold hyper_running in H_not_hyp_run.
  destruct H as (H_act, H').
  rewrite H_act in H_not_hyp_run; elim H_not_hyp_run; auto.

  (* New untrusted (hyper is running). *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold pre_action, pre_new_untrusted, hyper_running in H_pre.
  destruct H_pre as (H_os_untrusted, (ma', H_hyp_map, (H_ptype, (H_ma'_owner, (H_os_acc, (H_hcall, H_activity)))))).
  unfold hyper_running in H_not_hyp_run.
  rewrite H_activity in H_not_hyp_run; elim H_not_hyp_run; auto.

  (* New trusted *)
  apply lemma_new_trusted_isolation with s v p; auto.

  (* New hyper (hyper is running). *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold pre_action, pre_new_hyper, hyper_running in H_pre.
  destruct H_pre as (H_waitning, H').
  unfold hyper_running in H_not_hyp_run.
  rewrite H_waitning in H_not_hyp_run; elim H_not_hyp_run; auto.


  (* Del untrusted (hyper is running). *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold pre_action, pre_del_untrusted, hyper_running in H_pre.
  destruct H_pre as (H_os_untrtusted, (H_os_hcall, (system, H_system_map, (p2m, H_hyp_mapping,
    (pt_addr, H_pt_addr_map, (H_pre_all_del, (H_activity, H_accessible))))))).
  unfold hyper_running in H_not_hyp_run.
  rewrite H_activity in H_not_hyp_run; elim H_not_hyp_run; auto.

  (* Del trusted *)
  apply lemma_del_trusted_isolation with s v; auto.

  (* Del hyper (hyper is running). *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold pre_action, pre_del_hyper, pre_all_del, hyper_running in H_pre.
  destruct H_pre as (pt_addr, H_curr_map, (H_pre', (H_activity, H_not_accessible))). 
  unfold hyper_running in H_not_hyp_run.
  rewrite H_activity in H_not_hyp_run; elim H_not_hyp_run; auto.
  
  (* Switch (hyper is running). *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold pre_action, pre_switch, hyper_running in H_pre.
  destruct H_pre as (H_activity, H_pre').
  unfold hyper_running in H_not_hyp_run.
  rewrite H_activity in H_not_hyp_run; elim H_not_hyp_run; auto.

  (* LSwitch trusted *)
  apply lemma_lswitch_isolation with s (st_active_os s) p (Lswitch_trusted p); auto.

  (* LSwitch untrusted *)
  apply lemma_lswitch_isolation with s o p (Lswitch_untrusted o p); auto.

  (* Hcall (memory does not change). *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold post_action, post_hcall in H_post.
  destruct H_post as (aos, H_aos_map, H_s').
  rewrite H_s'; auto.

  (* Ret ctrl (memory does not change). *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold post_action, post_ret_ctrl in H_post.
  unfold state_components_assign, state_components_update in H_post.
  destruct H_post with System_Memory as [((H_triv1, H_triv2), H_eq_comp) | H_abs].
  unfold state_component_interp in H_eq_comp.
  unfold content_of_address.
  rewrite <- H_eq_comp.
  assumption.
  destruct H_abs as (H_abs, (H_exec_mode, H_activity)).
  cut False; [contradiction | (apply H_abs; split; discriminate)].

  (* Chmod (hyper is running). *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold pre_action, pre_chmod, hyper_running in H_pre.
  destruct H_pre as (H_activity, H_hcall).
  unfold hyper_running in H_not_hyp_run.
  rewrite H_activity in H_not_hyp_run; elim H_not_hyp_run; auto.

  (* Page pin trusted *)
  apply lemma_page_pin_trusted_isolation with s p p0; auto.

  (* Page unpin trusted *)
  apply lemma_page_unpin_trusted_isolation with s p; auto.

  (* Page pin untrusted (hyper is running). *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold pre_action, pre_page_pin_untrusted, pre_all_pin, hyper_running in H_pre.
  destruct H_pre as (H_ctxt, (H_hcall, (H_activity, H_pre'))).
  unfold hyper_running in H_not_hyp_run.
  rewrite H_activity in H_not_hyp_run; elim H_not_hyp_run; auto.

  (* Page unpin untrusted (hyper is running). *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold pre_action, pre_page_unpin_untrusted, pre_all_unpin, hyper_running in H_pre.
  destruct H_pre as (H_ctxt, (H_hcall, (H_activity, H_pre'))).
  unfold hyper_running in H_not_hyp_run.
  rewrite H_activity in H_not_hyp_run; elim H_not_hyp_run; auto.

Qed.
  

(* Write isolation - other formalization
   Write isolation captures the intuition that an OS cannot modify memory that
   it does not own.
*)

Definition write_isolation (s s':state) : Prop :=
 forall (a:action),
       valid_state ctxt s ->
       ~(hyper_running s) ->
       valid_exec ctxt s a s' ->     
       forall (ma:madd) (pg:page) (osi:os_ident), 
          content_of_address s ma pg ->
          ((owner_of_page pg osi /\ osi <> (st_active_os s)) \/
          page_owned_by pg = Hyp_PO) ->
          content_of_address s' ma pg.


Theorem write_isolation_proof : forall (s s':state), write_isolation s s'.
Proof.
  intros s s' a H_s_valid H_not_hyp_run H_valid_exec ma pg osi H_pg_map H_pg_owner.
  generalize (write_isolation_proof' s s' a H_s_valid H_not_hyp_run H_valid_exec ma pg H_pg_map); intro H_write_isolation.
  apply H_write_isolation.
  unfold owner_of_page; unfold owner_of_page in H_pg_owner; unfold not; intro H_aos.
  rewrite <- H_aos in H_pg_owner.
  destruct H_pg_owner as [(H_aos1,H_aos2) | H_hyp]; try discriminate.
  inversion H_aos1.
  rewrite H0 in H_aos2; elim H_aos; auto.
  unfold owner_of_page; unfold owner_of_page in H_pg_owner; unfold not; intro H_aos.
  rewrite H_aos in H_pg_owner.
  destruct H_pg_owner as [(H_aos1,H_aos2) | H_hyp]; try discriminate.
Qed.

End Write_Isolation.